@@ -364,7 +364,7 @@ xf86ProcessActionEvent(ActionEvent action, void *arg)
     case ACTION_SWITCHSCREEN:
 	if (VTSwitchEnabled && !xf86Info.dontVTSwitch && arg) {
 	    int vtno = *((int *) arg);
-#ifdef SCO
+#if defined(__SCO__) || defined(__UNIXWARE__)
 	    vtno--;
 #endif
 #if defined(QNX4)
@@ -377,12 +377,13 @@ xf86ProcessActionEvent(ActionEvent action, void *arg)
 	break;
     case ACTION_SWITCHSCREEN_NEXT:
 	if (VTSwitchEnabled && !xf86Info.dontVTSwitch) {
-#if defined(SCO) /* Shouldn't this be true for (sun) && (i386) && (SVR4) ? */
+/* Shouldn't this be true for (sun) && (i386) && (SVR4) ? */
+#if defined(__SCO__) || defined(__UNIXWARE__)
 	    if (ioctl(xf86Info.consoleFd, VT_ACTIVATE, xf86Info.vtno) < 0)
 #else
 	    if (ioctl(xf86Info.consoleFd, VT_ACTIVATE, xf86Info.vtno + 1) < 0)
 #endif
-#if defined (SCO) || (defined(sun) && defined (i386) && defined (SVR4))
+#if defined (__SCO__) || (defined(sun) && defined (i386) && defined (SVR4)) || defined(__UNIXWARE__)
 		if (ioctl(xf86Info.consoleFd, VT_ACTIVATE, 0) < 0)
 #else
 		if (ioctl(xf86Info.consoleFd, VT_ACTIVATE, 1) < 0)
@@ -784,7 +785,7 @@ special:
 	break;
 #endif
 
-#if defined(linux) || (defined(CSRG_BASED) && (defined(SYSCONS_SUPPORT) || defined(PCVT_SUPPORT) || defined(WSCONS_SUPPORT))) || defined(SCO)
+#if defined(linux) || (defined(CSRG_BASED) && (defined(SYSCONS_SUPPORT) || defined(PCVT_SUPPORT) || defined(WSCONS_SUPPORT))) || defined(__SCO__) || defined(__UNIXWARE__)
 	/*
 	 * Under Linux, the raw keycodes are consumed before the kernel
 	 * does any processing on them, so we must emulate the vt switching
@@ -923,7 +924,7 @@ special:
 
 #endif /* USE_VT_SYSREQ */
 
-#ifdef SCO
+#ifdef __SCO__
     /*
      *	With the console in raw mode, SCO will not switch consoles,
      *	you get around this by activating the next console along, if
@@ -938,7 +939,7 @@ special:
 	xf86ProcessActionEvent(ACTION_SWITCHSCREEN_NEXT, NULL);
       return;
     }
-#endif /* SCO */
+#endif /* __SCO__ */
 #ifdef XKB
     }
 #endif
@@ -1104,8 +1105,9 @@ special:
 Bool
 xf86CommonSpecialKey(int key, Bool down, int modifiers)
 {
-  if ((ModifierIsSet(ControlMask | AltMask)) ||
-      (ModifierIsSet(ControlMask | AltLangMask))) {
+  if ((!ModifierIsSet(ShiftMask)) &&
+      (((ModifierIsSet(ControlMask | AltMask)) ||
+        (ModifierIsSet(ControlMask | AltLangMask))))) {
       switch (key) {
 	
       case KEY_BackSpace:
@@ -1299,6 +1301,10 @@ static __inline__ void xorg_backtrace(void)
 
 #else /* not glibc or glibc < 2.1 */
 
+# if defined(sun) && defined(__SVR4)
+#  define HAVE_PSTACK
+# endif
+
 # if defined(HAVE_WALKCONTEXT) /* Solaris 9 & later */
 
 # include <ucontext.h>
@@ -1357,14 +1363,91 @@ static int xorg_backtrace_frame(uintptr_t pc, int signo, void *arg)
     
     return 0;
 }
+# endif /* HAVE_WALKCONTEXT */
 
-static __inline__ void xorg_backtrace(void) { 
-    ucontext_t u;
-    int depth = 1;
+# ifdef HAVE_PSTACK
+static int xorg_backtrace_pstack(void) {
+    pid_t kidpid;
+    int pipefd[2];
 
-    if (getcontext(&u) == 0)
-	walkcontext(&u, xorg_backtrace_frame, &depth);
+    if (pipe(pipefd) != 0) {
+	return -1;
+    }
+
+    kidpid = fork1();
+
+    if (kidpid == -1) {
+	/* ERROR */
+	return -1;
+    } else if (kidpid == 0) {
+	/* CHILD */
+	char parent[16];
+	
+	seteuid(0);
+	close(STDIN_FILENO);
+	close(STDOUT_FILENO);
+	dup2(pipefd[1],STDOUT_FILENO);
+	closefrom(STDERR_FILENO);
+
+	snprintf(parent, sizeof(parent), "%d", getppid());
+	execle("/usr/bin/pstack", "pstack", parent, NULL);
+	exit(1);
+    } else {
+	/* PARENT */
+	char btline[256];
+	int kidstat;
+	int bytesread;
+	int done = 0;
+	
+	close(pipefd[1]);
+
+	while (!done) {
+	    bytesread = read(pipefd[0], btline, sizeof(btline) - 1);
+	    
+	    if (bytesread > 0) {
+		btline[bytesread] = 0;
+		ErrorF("%s", btline);
+	    }
+	    else if ((bytesread < 0) ||
+		     ((errno != EINTR) && (errno != EAGAIN)))
+		done = 1;
+	}
+	close(pipefd[0]);
+	waitpid(kidpid, &kidstat, 0);
+	if (kidstat != 0)
+	    return -1;
+    }
+    return 0;
 }
+# endif /* HAVE_PSTACK */
+
+
+# if defined(HAVE_PSTACK) || defined(HAVE_WALKCONTEXT)
+
+static __inline__ void xorg_backtrace(void) {
+
+    ErrorF("\nBacktrace:\n");
+    
+#  ifdef HAVE_PSTACK    
+/* First try fork/exec of pstack - otherwise fall back to walkcontext
+   pstack is preferred since it can print names of non-exported functions */
+    
+    if (xorg_backtrace_pstack() < 0)
+#  endif	
+    {
+#  ifdef HAVE_WALKCONTEXT
+	ucontext_t u;
+	int depth = 1;
+	
+	if (getcontext(&u) == 0)
+	    walkcontext(&u, xorg_backtrace_frame, &depth);
+	else
+#  endif
+	    Error("Failed to get backtrace info");
+    }
+    ErrorF("\n");	
+}
+
 # else
 
 /* Default fallback if we can't find any way to get a backtrace */
